<?php
/**
 * @brief		Background Task
 * @author		<a href='https://www.invisioncommunity.com'>Invision Power Services, Inc.</a>
 * @copyright	(c) Invision Power Services, Inc.
 * @license		https://www.invisioncommunity.com/legal/standards/
 * @package		Invision Community
 * @subpackage	Chatbox+
 * @since		05 Mar 2019
 */

namespace IPS\chatbox\extensions\core\Queue;

/* To prevent PHP errors (extending class does not exist) revealing path */
if ( !\defined( '\IPS\SUITE_UNIQUE_KEY' ) )
{
	header( ( isset( $_SERVER['SERVER_PROTOCOL'] ) ? $_SERVER['SERVER_PROTOCOL'] : 'HTTP/1.0' ) . ' 403 Forbidden' );
	exit;
}

/**
 * Background Task
 */
class _deleteMsg
{

	/**
	 * @brief Number of topics to build per cycle
	 */
	public $rebuild = 20;
	
	/**
	 * Parse data before queuing
	 *
	 * @param	array	$data
	 * @return	array
	 */
	public function preQueueData( $data )
	{
		# For messages in conversations
		if ( $data['convo'] > 0 )
		{
			$table = 'chatbox_conversations_posts';
			
			if ( $data['convo'] > 0 )
			{
				$where[] = array( 'con=?', $data['convo'] ); 
			}
			if ( $data['user'] )
			{
				$where[] = array( 'member_id=?', $data['user'] ); 
			}			
		}
		
		# For messages in rooms
		else
		{
			$table = 'chatbox_messages';
			
			if ( $data['room'] > 0 )
			{
				$where[] = array( '( room=? OR FIND_IN_SET( ?, in_multiple_rooms ) )', $data['room'], $data['room'] ); 
			}
			if ( $data['user'] )
			{
				$key = explode( "@", $data['user'] );
				
				if ( $key[0] == 'user' )
				{
					$where[] = array( 'member_id=?', $key[1] ); 
				}
				else
				{
					$where[] = array( 'guestName=?', $key[1] ); 
				}
			}
		}
		
		# Count
		try
		{
			$data['count']		= \IPS\Db::i()->select( 'MAX(id)', $table, $where )->first();
			$data['realCount']	= \IPS\Db::i()->select( 'COUNT(*)', $table, $where )->first();
		}
		catch( \Exception $ex )
		{
			return NULL;
		}
		
		if( $data['count'] == 0 or $data['realCount'] == 0 )
		{
			return NULL;
		}

		$data['indexed'] = 0;
		
		return $data;
	}

	/**
	 * Run Background Task
	 *
	 * @param	mixed						$data	Data as it was passed to \IPS\Task::queue()
	 * @param	int							$offset	Offset
	 * @return	int							New offset
	 * @throws	\IPS\Task\Queue\OutOfRangeException	Indicates offset doesn't exist and thus task is complete
	 */
	public function run( $data, $offset )
	{
		$count = 0;
		
		# Delete messages in conversations
		if ( $data['convo'] > 0 )
		{
			if ( $data['convo'] > 0 )
			{
				$where[] = array( 'con=?', $data['convo'] ); 
			}
			if ( $data['user'] )
			{
				$where[] = array( 'member_id=?', $data['user'] ); 
			}
			
			$select = \IPS\Db::i()->select( '*', 'chatbox_conversations_posts', $where, 'id', array( 0, $this->rebuild ) );
			
			foreach( $select as $chat )
			{
				$delete[] = $chat['id'];
				$data['indexed']++;
				$last = $chat['id'];				
			}	
			
			if ( \count( $delete ) > 0 )
			{
				\IPS\Db::i()->delete( 'chatbox_conversations_posts', array( 'id IN (?)', implode( ",", $delete ) ) );
			}			
		}
		
		# Delete messages in rooms
		else
		{
			if ( $data['room'] > 0 ) // msg in multiple rooms
			{
				$where[] = array( '( room=? OR FIND_IN_SET( ?, in_multiple_rooms ) )', $data['room'], $data['room'] ); 
				
				if ( $data['user'] )
				{
					$key = explode( "@", $data['user'] );
					$where[] = ( $key[0] == 'user' ) ? array( 'member_id=?', $key[1] ) : array( 'guestName=?', $key[1] ); 
				}
				
				$select = \IPS\Db::i()->select( '*', 'chatbox_messages', $where, 'id', array( 0, $this->rebuild ) );
				
				foreach( $select as $chat )
				{
					if ( $chat['room'] > 0 || $chat['in_multiple_rooms'] === (string) $data['room'] )
					{
						$delete[] = $chat['id'];
					}
					else
					{
						$r = explode( ",", $chat['in_multiple_rooms'] );
						$rooms = array_diff( $r, [ $data['room'] ] );
						\IPS\Db::i()->update( 'chatbox_messages', array( 'in_multiple_rooms' => implode( ",", $rooms ) ), array( 'id=?', $chat['id'] ) );
					}
					
					$data['indexed']++;
					$last = $chat['id'];				
				}
			}
			else // msg in 1 room only
			{
				$select = \IPS\Db::i()->select( '*', 'chatbox_messages', $where, 'id', array( 0, $this->rebuild ) );
				
				foreach( $select as $chat )
				{
					$delete[] = $chat['id'];
					$data['indexed']++;
					$last = $chat['id'];				
				}
			}
			
			if ( \count( $delete ) > 0 )
			{
				\IPS\Db::i()->delete( 'chatbox_messages', array( 'id IN (?)', implode( ",", $delete ) ) );
			}
		}
		
		# No last
		if( $last === NULL )
		{
			throw new \IPS\Task\Queue\OutOfRangeException;
		}
		
		# Done
		return $last;
	}
	
	/**
	 * Get Progress
	 *
	 * @param	mixed					$data	Data as it was passed to \IPS\Task::queue()
	 * @param	int						$offset	Offset
	 * @return	array( 'text' => 'Doing something...', 'complete' => 50 )	Text explaining task and percentage complete
	 * @throws	\OutOfRangeException	Indicates offset doesn't exist and thus task is complete
	 */
	public function getProgress( $data, $offset )
	{
		return array( 'text' => \IPS\Member::loggedIn()->language()->addToStack('chatbox_queue_delete_msg'), 'complete' => ( $data['realCount'] * $data['indexed'] ) > 0 ? round( ( $data['realCount'] * $data['indexed'] ) * 100, 2 ) : 0 );
	}

	/**
	 * Perform post-completion processing
	 *
	 * @param	array	$data
	 * @return	void
	 */
	public function postComplete( $data )
	{

	}
}