<?php

namespace IPS\chatbox\Rooms;

/* To prevent PHP errors (extending class does not exist) revealing path */
if ( !defined( '\IPS\SUITE_UNIQUE_KEY' ) )
{
	header( ( isset( $_SERVER['SERVER_PROTOCOL'] ) ? $_SERVER['SERVER_PROTOCOL'] : 'HTTP/1.0' ) . ' 403 Forbidden' );
	exit;
}

/**
 * RSS Exports
 */
class _Room extends \IPS\Node\Model
{
	use \IPS\Content\ClubContainer;
	
	/**
	 * @brief	[ActiveRecord] Multiton Store
	 */
	protected static $multitons;
	
	/**
	 * @brief	[ActiveRecord] Database Table
	 */
	public static $databaseTable = 'chatbox_rooms';
	
	/**
	 * @brief	[ActiveRecord]	Database Prefix
	 */
	public static $databasePrefix = 'room_';
			
	/**
	 * @brief	[Node] Order Database Column
	 */
	public static $databaseColumnOrder = 'position';
		
	/**
	 * @brief	[Node] Node Title
	 */
	public static $nodeTitle = 'chatbox_rooms';
	
	/**
	 * @brief	Content Item Class
	 */
	public static $contentItemClass = 'IPS\chatbox\Rooms\Chat';
	
	/**
	 * @brief	[Node] Title prefix.  If specified, will look for a language key with "{$key}_title" as the key
	 */
	public static $titleLangPrefix = 'chatbox_room_';
	
	/**
	 * @brief	[Node] Description suffix.  If specified, will look for a language key with "{$titleLangPrefix}_{$id}_{$descriptionLangSuffix}" as the key
	 */
	public static $descriptionLangSuffix = '_desc';
	
	/**
	 * @brief	[Node] Enabled/Disabled Column
	 */
	public static $databaseColumnEnabledDisabled = 'enabled';
	
	/**
	 * @brief	[Node] ACP Restrictions
	 * @code
	 	array(
	 		'app'		=> 'core',				// The application key which holds the restrictrions
	 		'module'	=> 'foo',				// The module key which holds the restrictions
	 		'map'		=> array(				// [Optional] The key for each restriction - can alternatively use "prefix"
	 			'add'			=> 'foo_add',
	 			'edit'			=> 'foo_edit',
	 			'permissions'	=> 'foo_perms',
	 			'delete'		=> 'foo_delete'
	 		),
	 		'all'		=> 'foo_manage',		// [Optional] The key to use for any restriction not provided in the map (only needed if not providing all 4)
	 		'prefix'	=> 'foo_',				// [Optional] Rather than specifying each  key in the map, you can specify a prefix, and it will automatically look for restrictions with the key "[prefix]_add/edit/permissions/delete"
	 * @encode
	 */
	protected static $restrictions = array(
		'app'		=> 'chatbox',
		'module'	=> 'chatbox',
		'all'	 	=> 'rooms_manage',
	);
	
	/**
	 * @brief	The map of permission columns
	 */
	public static $permissionMap = array(
			'view' 				=> 'view',
			'read'				=> 2,
	);	
	
	/**
	 * @brief	[Node] URL Base
	 */
	public static $urlBase = 'app=chatbox&module=chatbox&controller=room&id=';
	
	/**
	 * @brief	[Node] SEO Template
	 */
	public static $urlTemplate = 'chatbox_room';
	
	/** 
	 * @brief	[Node] App for permission index
	 */
	public static $permApp = 'chatbox';
	
	/** 
	 * @brief	[Node] Type for permission index
	 */
	public static $permType = 'room';
	
	/**
	 * @brief	[Node] Moderator Permission
	 */
	public static $modPerm = 'chatbox_rooms';
		
	/**
	 * @brief	SEO Title Column
	 */
	public static $seoTitleColumn = 'seo_name';
	
	/**
	 * Get URL for doing stuff
	 *
	 * @param	string|NULL		$action		Action
	 * @return	\IPS\Http\Url
	 */
	public function url( $action=null )
	{
		$url = parent::url( $action );
		
		if ( $action )
		{
			return $url->setQueryString( 'do', $action );
		}
		return $url;
	}
	
	/**
	 * [Node] Does the currently logged in user have permission to edit permissions for this node?
	 *
	 * @return	bool
	 */
	public function canManagePermissions()
	{
		return false;
	}
	
	/**
	 * [Node] Does the currently logged in user have permission to mass move/delete content in this node?
	 *
	 * @return	bool
	 */
	public function canMassManageContent()
	{
		return FALSE;
	}
	
	/**
	 * Get data store
	 *
	 * @return	array
	 * @note	Note that all records are returned, even disabled report rules. Enable status needs to be checked in userland code when appropriate.
	 */
	public static function getStore()
	{
		if ( !isset( \IPS\Data\Store::i()->chatbox_rooms ) )
		{
			\IPS\Data\Store::i()->chatbox_rooms = iterator_to_array( \IPS\Db::i()->select( '*', static::$databaseTable, NULL, "room_id ASC" )->setKeyField( 'room_id' ) );
		}
		
		return \IPS\Data\Store::i()->chatbox_rooms;
	}

	/**
	 * @brief	[ActiveRecord] Attempt to load from cache
	 * @note	If this is set to TRUE you MUST define a getStore() method to return the objects from cache
	 */
	protected static $loadFromCache = TRUE;
	
	/**
	 * @brief	[ActiveRecord] Caches
	 * @note	Defined cache keys will be cleared automatically as needed
	 */
	protected $caches = array( 'chatbox_rooms' );
	
	/**
	 * Load Record
	 *
	 * @see		\IPS\Db::build
	 * @param	int|string	$id					ID
	 * @param	string		$idField			The database column that the $id parameter pertains to (NULL will use static::$databaseColumnId)
	 * @param	mixed		$extraWhereClause	Additional where clause(s) (see \IPS\Db::build for details)
	 * @return	static
	 * @throws	\InvalidArgumentException
	 * @throws	\OutOfRangeException
	 */
	public static function load( $id, $idField=NULL, $extraWhereClause=NULL )
	{
		return parent::load( $id, $idField, $extraWhereClause );
	}
	
	/**
	 * [Node] Add/Edit Form
	 *
	 * @param	\IPS\Helpers\Form	$form	The form
	 * @return	void
	 */
	public function form( &$form )
	{
		$opts = $this->options();

		# General settings
		$form->addTab( 'room_tab_general' );
		$form->add( new \IPS\Helpers\Form\Translatable( 'room_name', NULL, TRUE, array( 'app' => 'chatbox', 'key' => ( $this->_id ) ? "chatbox_room_{$this->_id}" : NULL ) ) );
		$form->add( new \IPS\Helpers\Form\Translatable( 'room_description', NULL, FALSE, array( 'app' => 'chatbox', 'key' => ( $this->_id ) ? "chatbox_room_{$this->_id}_description" : NULL, 'textArea' => TRUE ) ) );
		$form->add( new \IPS\Helpers\Form\YesNo( 'room_show_rules', $this->show_rules, FALSE, array( 'togglesOn' => array('room_rules') ) ) );
		$form->add( new \IPS\Helpers\Form\Translatable( 'room_rules', NULL, FALSE, array( 'app' => 'chatbox', 'key' => ( $this->id ? "chatbox_room_{$this->id}_rules" : NULL ), 'editor' => array( 'app' => 'chatbox', 'key' => 'Room', 'autoSaveKey' => ( $this->id ? "chatbox-room-{$this->id}-rules" : "chatbox-room-new-rules" ), 'attachIds' => $this->id ? array( $this->id, NULL, 'chatbox_rules' ) : NULL, 'minimize' => 'chatbox_rules_placeholder' ) ), NULL, NULL, NULL, 'room_rules' ) );
		$form->add( new \IPS\Helpers\Form\Text( 'room_icon', $this->icon ? $this->icon : 'comments' ) );

		$form->add( new \IPS\Helpers\Form\Upload( 'room_notifile', $this->notifile ? \IPS\File::get( 'chatbox_Nofifile', $this->notifile ) : NULL, FALSE, array( 
			'allowedFileTypes'	=> array( 'mp3' ),
			'maxFileSize'		=> 1,
			'storageExtension' 	=> 'chatbox_Nofifile',
		), NULL, NULL, NULL, 'room_notifile' ) );

		$form->add( new \IPS\Helpers\Form\Select( 'chatbox_style', $opts ? $opts['style'] : 'standard', FALSE, array( 'options' => \IPS\Application::load('chatbox')->styles() ) ) );	
		foreach( \IPS\Application::load('chatbox')->skins() as $bg )
		{
			$skinClass = $bg != 'white' ? $bg . '_box' : 'ipsType_sectionTitle white_box';
			$skinSelect[ $bg ] = "<div class='ipsBox {$skinClass}' style='height: 20px; width: 80px;'></div>";
		}		
		$form->add( new \IPS\Helpers\Form\Radio( 'chatbox_skin', $opts['skin'] ? $opts['skin'] : 'white', FALSE, array( 'options' => $skinSelect, 'parse' => 'raw' ) ) );		
		$form->add( new \IPS\Helpers\Form\YesNo( 'chatbox_allow_select_skins', $opts['allow_select_skins'], FALSE, array(), NULL, NULL, NULL, 'chatbox_allow_select_skins' ) );
		$form->add( new \IPS\Helpers\Form\Number( 'chatbox_height', $opts['height'] ? \intval( $opts['height'] ) : 420, TRUE, array( 'min' => 200 ) ) );
		
		# Chat
		$form->addTab( 'room_tab_chat' );
		$form->add( new \IPS\Helpers\Form\Select( 'chatbox_sort', $opts ? $opts['sort'] : 'asc', FALSE, array( 'options' => array(
			'asc'	=> 'chatbox_asc_sort',
			'desc'	=> 'chatbox_desc_sort'
		) ) ) );		

		$form->add( new \IPS\Helpers\Form\Number( 'chatbox_limit', $opts['limit'] ? \intval( $opts['limit'] ) : 20, TRUE, array( 'min' => 5 ) ) );
		$form->add( new \IPS\Helpers\Form\Number( 'chatbox_interval', $opts['interval'] ? \intval( $opts['interval'] ) : 10000, TRUE, array( 'min' => 2000 ) ) );
		$form->add( new \IPS\Helpers\Form\Number( 'chatbox_flood', $opts['flood'] ? \intval( $opts['flood'] ) : 5, FALSE, array( 'unlimited' => 0, 'unlimitedLang' => 'none' ) ) );
		$form->add( new \IPS\Helpers\Form\Number( 'chatbox_length', isset( $opts['length'] ) ? \intval( $opts['length'] ) : 300, FALSE, array( 'unlimited' => 0, 'unlimitedLang' => 'unlimited' ) ) );
		$form->add( new \IPS\Helpers\Form\YesNo( 'chatbox_autoload', $opts['autoload'], FALSE ) );
		$form->add( new \IPS\Helpers\Form\YesNo( 'chatbox_showUserList', $opts['showUserList'], FALSE ) );
		$form->add( new \IPS\Helpers\Form\YesNo( 'chatbox_auto_guest', $opts['auto_guest'], FALSE ) );
		
		# Input
		$form->addTab('chatbox_tab_input_media');
		$form->addHeader('chatbox_tab_media');
		$form->add( new \IPS\Helpers\Form\YesNo( 'chatbox_musicbox', $opts['musicbox'], FALSE, array(), NULL, NULL, NULL, 'chatbox_musicbox' ) );

		$form->addHeader('chatbox_tab_input');
		$form->add( new \IPS\Helpers\Form\YesNo( 'chatbox_use_editor', $opts['use_editor'], FALSE, array(
			'togglesOff' => array( 'chatbox_allow_img', 'chatbox_allow_video', 'chatbox_giphy', 'chatbox_guggy' ),
			'togglesOn'	 => array( 'chatbox_enter_editor', 'chatbox_editor_upload' )
		) ) );	
		$form->add( new \IPS\Helpers\Form\YesNo( 'chatbox_editor_upload', $opts['editor_upload'], FALSE, array(), NULL, NULL, NULL, 'chatbox_editor_upload' ) );
		$form->add( new \IPS\Helpers\Form\YesNo( 'chatbox_enter_editor', $opts['enter_editor'], FALSE, array(), NULL, NULL, NULL, 'chatbox_enter_editor' ) );
		
		$form->add( new \IPS\Helpers\Form\YesNo( 'chatbox_allow_img', $opts['allow_img'], FALSE, array(), NULL, NULL, NULL, 'chatbox_allow_img' ) );
		$form->add( new \IPS\Helpers\Form\YesNo( 'chatbox_allow_video', $opts['allow_video'], FALSE, array(), NULL, NULL, NULL, 'chatbox_allow_video' ) );
		$form->add( new \IPS\Helpers\Form\Select( 'chatbox_giphy', $opts['giphy'] ? $opts['giphy'] : 'no', FALSE, array( 'options' => array( 'no' => 'chatbox_giphy_no', 'button' => 'chatbox_giphy_button', 'cmd' => 'chatbox_giphy_cmd', 'both' => 'chatbox_giphy_both' ) ), NULL, NULL, NULL, 'chatbox_giphy' ) );	
		$form->add( new \IPS\Helpers\Form\Select( 'chatbox_guggy', $opts['guggy'] ? $opts['guggy'] : 'no', FALSE, array( 'options' => array( 'no' => 'chatbox_guggy_no', 'button' => 'chatbox_guggy_button', 'cmd' => 'chatbox_guggy_cmd', 'both' => 'chatbox_guggy_both' ) ), NULL, NULL, NULL, 'chatbox_guggy' ) );	
		
		# System messages
		$form->addTab('chatbox_tab_auto_msg');
		
		$form->addHeader('chatbox_header_sysMSG');
		$form->add( new \IPS\Helpers\Form\YesNo( 'chatbox_announcement_on', $opts['announcement_on'] ) );
		$form->add( new \IPS\Helpers\Form\YesNo( 'chatbox_online_on', $opts['online_on'] ) );
		
		$form->addHeader('sprofile_header_botMSG');
		$form->add( new \IPS\Helpers\Form\YesNo( 'room_auto_msg_on', $this->auto_msg_on, FALSE, array(
			'togglesOn' => array( 'room_auto_msg', 'room_auto_msg_user' )
		) ) );
		$form->add( new \IPS\Helpers\Form\Member( 'room_auto_msg_user', $this->auto_msg_user > 0 ? \IPS\Member::load( $this->auto_msg_user ) : NULL, TRUE, array( 
			'multiple' 	=> 1
		), NULL, NULL, NULL, 'room_auto_msg_user' ) );
		$form->add( new \IPS\Helpers\Form\Stack( 'room_auto_msg', $this->auto_msg ? json_decode( $this->auto_msg, TRUE ) : array(), TRUE, array( 
			'stackFieldType' 	=> 'IPS\chatbox\Form\AutoMSG',
			'maxItems'			=> 5,
		), NULL, NULL, NULL, 'room_auto_msg' ) );
		
		
		# Permissions
		$form->addTab( 'room_tab_permissions' );
		$groups = \IPS\Member\Group::groups();	
		$form->add( new \IPS\Helpers\Form\Select( 'room_can_view', array_filter( explode( ',', $this->can_view ) ), FALSE, array( 'options' => $groups, 'parse' => 'normal', 'multiple' => TRUE ) ) );		
		$form->add( new \IPS\Helpers\Form\Select( 'room_can_chat', array_filter( explode( ',', $this->can_chat ) ), FALSE, array( 'options' => $groups, 'parse' => 'normal', 'multiple' => TRUE ) ) );
		$form->add( new \IPS\Helpers\Form\Select( 'room_can_see_online', array_filter( explode( ',', $this->can_see_online ) ), FALSE, array( 'options' => $groups, 'parse' => 'normal', 'multiple' => TRUE ) ) );
		$form->add( new \IPS\Helpers\Form\Select( 'room_can_manage', array_filter( explode( ',', $this->can_manage ) ), FALSE, array( 'options' => $groups, 'parse' => 'normal', 'multiple' => TRUE ) ) );
	}
	
	/**
	 * [Node] Format form values from add/edit form for save
	 *
	 * @param	array	$values	Values from the form
	 * @return	array
	 */
	public function formatFormValues( $values )
	{
		if ( !$this->_id )
		{
			$this->save();
			\IPS\File::claimAttachments( 'chatbox-room-new-rules', $this->id, NULL, 'chatbox_rules' );
		}
		
		# Custom Language Strings
		if ( array_key_exists( 'room_name', $values ) )
		{
			\IPS\Lang::saveCustom( 'chatbox', "chatbox_room_{$this->_id}", $values['room_name'] );
			
			if ( \is_array( $values['room_name'] ) )
			{
				reset( $values['room_name'] );
				$this->seo_name = \IPS\Http\Url\Friendly::seoTitle( $values['room_name'][ key( $values['room_name'] ) ] );
			}
			else
			{
				$this->seo_name = \IPS\Http\Url\Friendly::seoTitle( $values['room_name'] );
			}
			
			unset( $values['room_name'] );
		}
		
		if ( array_key_exists( 'room_description', $values ) )
		{
			\IPS\Lang::saveCustom( 'chatbox', "chatbox_room_{$this->_id}_description", $values['room_description'] );	
			unset( $values['room_description'] );
		}
		
		# Rules
		if ( array_key_exists( 'room_rules', $values ) )
		{
			\IPS\Lang::saveCustom( 'chatbox', "chatbox_room_{$this->_id}_rules", $values['room_rules'] );
			unset( $values['room_rules'] );
		}
		
		# Sound
		$this->notifile = $values['room_notifile'] ? (string) $values['room_notifile'] : NULL;
		unset( $values['room_notifile'] );
		
		# Auto message
		$this->auto_msg_user = $values['room_auto_msg_user']->member_id;
		unset( $values['room_auto_msg_user'] );
		
		foreach( $values['room_auto_msg'] as $stack )
		{
			$stack['time'] = ( $stack['time'] >= 10 ) ? $stack['time'] : 10;
			$stack['last'] = time();
			if ( $stack['msg'] )
			{
				$stackData[] = $stack;
			}				
		}

		$this->auto_msg = \count( $stackData ) > 0 ? json_encode( $stackData ) : NULL;
		unset( $values['room_auto_msg'] );
		
		# Other options
		foreach( \array_keys( $values ) as $key )
		{
			if ( mb_strpos( $key, 'chatbox_' ) !== FALSE )
			{
				$name = str_replace( 'chatbox_', '', $key );
				$dataOpts[ $name ] = $values[ $key ];
				unset( $values[ $key ] );
			}
		}

		if ( \count( $dataOpts ) > 0 )
		{
			$this->options = json_encode( $dataOpts );
		}
		
		# Save
		$this->save();
		
		# Pass to parent
		return parent::formatFormValues( $values );
	}
	
	/*-------------------------------------------------------------------------*/
	// Options
	/*-------------------------------------------------------------------------*/		
	protected $_opts = NULL;
	
	public function options()
	{
		if ( $this->_opts[ $this->id ] !== NULL )
		{
			return $this->_opts[ $this->id ];
		}
		
		$data = array();
		
		if ( $club = $this->club() )
		{
			$data = json_decode( \IPS\Settings::i()->room_clubs_configs, TRUE );
			$sound = \IPS\Settings::i()->room_clubs_notifile;		
		}
		else
		{
			$data = json_decode( $this->options, true );
			$sound = $this->notifile;			
		}

		if ( \count( $data ) == 0 )
		{
			$data = array (
				'style' 				=> 'standard',
				'skin' 					=> 'white',
				'height' 				=> 300,
				'sort' 					=> 'asc',
				'limit' 				=> 20,
				'interval' 				=> 10000,
				'flood' 				=> 5,
				'length' 				=> 300,
			);
		}
		
		if ( $sound )
		{
			try
			{
				$data['soundfile'] = (string) \IPS\File::get( 'chatbox_Nofifile', $sound )->url;
			}
			catch( \Exception $ex ) { }			
		}

		$this->_opts[ $this->id ] = $data;
		return $this->_opts[ $this->id ];		
	}

	/*-------------------------------------------------------------------------*/
	// Auto messages
	/*-------------------------------------------------------------------------*/		
	protected $_autoMsg = NULL;
	
	public function autoMsgData()
	{
		if ( $this->auto_msg_on != 1  || !$this->auto_msg )
		{
			return NULL;
		}
		
		if ( $this->_autoMsg[ $this->id ] !== NULL )
		{
			return $this->_autoMsg[ $this->id ];
		}
		
		$mem = \IPS\Member::load( $this->auto_msg_user );
		
		$data['bot'] = array(
			'chatterID' 		=> $mem->member_id,
			'chatterName' 		=> $mem->name,
			'chatterKey' 		=> "user@" . $mem->member_id,
			'chatterPhoto' 		=> $mem->photo,
			'chatterNameFormat' => $mem->member_group_id ? \IPS\Member\Group::load( $mem->member_group_id )->formatName( $mem->name ) : $mem->name,
		);
		
		$data['messages'] = json_decode( $this->auto_msg, TRUE );
		$this->_autoMsg[ $this->id ] = $data;
		return $this->_autoMsg[ $this->id ];		
	}
	
	/**
	 * [ActiveRecord] Delete Record
	 *
	 * @return	void
	 */
	public function delete()
	{
		\IPS\File::unclaimAttachments( 'chatbox_Room', $this->id );
		
		// Delete uploaded sound
		try
		{
			\IPS\File::get( 'chatbox_Nofifile', $this->notifile )->delete();
		}
		catch( \Exception $ex ) { }
		
		// Delete language bits
		foreach ( array( 'room_name' => "chatbox_room_{$this->_id}", 'room_rules' => "chatbox_room_{$this->id}_rules" ) as $fieldKey => $langKey )
		{
			\IPS\Lang::deleteCustom( 'chatbox', $langKey );
		}
		
		// Delete cache
		$roomData = 'chatbox_room_' . $this->id;
		if ( isset( \IPS\Data\Store::i()->$roomData ) )
		{
			unset( \IPS\Data\Store::i()->$roomData );
		}
		if ( isset( \IPS\Data\Store::i()->chatbox_rooms_msg ) )
		{
			unset( \IPS\Data\Store::i()->chatbox_rooms_msg[ $this->id ] );
		}
		
		// Delete messages
		$this->cleanRoom();
		
		// Delete room
		parent::delete();
	}
	
	/**
	 * Delete all messages in room
	 *
	 * @return	void
	 */	
	public function cleanRoom()
	{
		\IPS\Task::queue( 'chatbox', 'deleteMsg', array( 'room' => $this->id ), 4 );
	}

	/* !Clubs */
	
	/**
	 * Get acp language string
	 *
	 * @return	string
	 */
	public static function clubAcpTitle()
	{
		return 'chatbox_room';
	}
	
	/**
	 * Set form for creating a node of this type in a club
	 *
	 * @param	\IPS\Helpers\Form	$form	Form object
	 * @return	void
	 */
	public function clubForm( \IPS\Helpers\Form $form )
	{
		$itemClass = static::$contentItemClass;

		$form->addTab( 'room_tab_general' );
		$form->add( new \IPS\Helpers\Form\Text( 'club_node_name', $this->_id ? $this->_title : NULL, TRUE, array( 'maxLength' => 255 ) ) );
		$form->add( new \IPS\Helpers\Form\YesNo( 'room_show_rules', $this->show_rules, FALSE, array( 'togglesOn' => array('room_rules') ) ) );
		$form->add( new \IPS\Helpers\Form\Translatable( 'room_rules', NULL, FALSE, array( 'app' => 'chatbox', 'key' => ( $this->id ? "chatbox_room_{$this->id}_rules" : NULL ), 'editor' => array( 'app' => 'chatbox', 'key' => 'Room', 'autoSaveKey' => ( $this->id ? "chatbox-room-{$this->id}-rules" : "chatbox-room-new-rules" ), 'attachIds' => $this->id ? array( $this->id, NULL, 'chatbox_rules' ) : NULL, 'minimize' => 'chatbox_rules_placeholder' ) ), NULL, NULL, NULL, 'room_rules' ) );
	}
	
	/**
	 * Class-specific routine when saving club form
	 *
	 * @param	\IPS\Member\Club	$club	The club
	 * @param	array				$values	Values
	 * @return	void
	 */
	public function _saveClubForm( \IPS\Member\Club $club, $values )
	{
		# Name
		if ( $values['club_node_name'] )
		{
			\IPS\Lang::saveCustom( 'chatbox', "chatbox_room_{$this->_id}", $values['club_node_name'] );
			$this->seo_name = \IPS\Http\Url\Friendly::seoTitle( $values['club_node_name'] );
		}		

		# Rules
		$this->show_rules = (int) $values['room_show_rules'];
		unset( $values['room_show_rules'] );
		
		if ( array_key_exists( 'room_rules', $values ) )
		{
			\IPS\Lang::saveCustom( 'chatbox', "chatbox_room_{$this->_id}_rules", $values['room_rules'] );
			
			unset( $values['room_rules'] );
		}

		if ( !$this->_id )
		{
			$this->save();
			\IPS\File::claimAttachments( 'chatbox-room-new-rules', $this->id, NULL, 'chatbox_rules' );
		}
	}
	
	/**
	 * Permissions
	 *
	 * @return	boolean
	 */	
	protected $_isBlocked = NULL;
	
	public function isBlocked()
	{
		if ( $this->_isBlocked !== NULL )
		{
			return $this->_isBlocked;
		}
		
		$data = array();
		
		if ( $this->blocked )
		{
			$data = json_decode( $this->blocked, true );
		}
		
		$this->_isBlocked = false;
		
		if ( !$this->_canManage() )
		{
			foreach( $data as $k => $v )
			{
				if ( $k == $this->userKey() )
				{
					$this->_isBlocked = true;
					break;
				}
			}
		}
		
		return $this->_isBlocked;
	}
	
	public function _canView( $member=NULL )
	{
		$member = $member ?: \IPS\Member::loggedIn();
		
		if ( $this->enabled != 1 )
		{
			return FALSE;
		}
		
		if ( $member->member_id > 0 && ( $member->restrict_post !== 0 || $member->temp_ban !== 0 ) )
		{
			return FALSE;
		}
		
		if ( $this->_canManage($member) OR $member->inGroup( explode( ",", $this->can_view ) ) )
		{
			return TRUE;		
		}
		
		if ( $club = $this->club() )
		{
			if ( $club->canRead() )
			{
				return TRUE;
			}
		}		

		return FALSE;	
	}

	public function _canChat( $member=NULL )
	{
		$member = $member ?: \IPS\Member::loggedIn();
		
		if ( $this->enabled != 1 )
		{
			return FALSE;
		}
		
		if ( $member->member_id > 0 && ( $member->restrict_post !== 0 || $member->temp_ban !== 0 ) )
		{
			return FALSE;
		}
		
		if ( $this->_canManage($member) OR $member->inGroup( explode( ",", $this->can_chat ) ) )
		{
			return TRUE;		
		}

		if ( $club = $this->club() )
		{
			if ( $club->canPost() )
			{
				return TRUE;
			}
		}
		
		return FALSE;		
	}	
	
	public function _canManage( $member=NULL )
	{
		$member = $member ?: \IPS\Member::loggedIn();
		
		if ( $club = $this->club() )
		{
			if ( $club->isLeader() || $club->isModerator() )
			{
				return TRUE;
			}
		}
				
		if ( $member->isAdmin() || $member->inGroup( explode( ",", $this->can_manage ) ) )
		{
			return TRUE;		
		}

		return FALSE;		
	}
	
	public function _canEdit( $chatterID=0, $member=NULL )
	{
		$member = $member ?: \IPS\Member::loggedIn();
		
		if ( $this->_canManage( $member ) || \IPS\chatbox\Rooms\Chat::modPermission( 'edit', NULL, $this ) )
		{
			return TRUE;		
		}

		if ( $chatterID > 0 && $member->member_id == $chatterID )
		{
			$perm = $member->_group['g_edit_posts'];
			return ( $perm == 1 OR \in_array( 'IPS\chatbox\Rooms\Chat', explode( ',', $perm ) ) ) ? TRUE : FALSE;			
		}
		
		return FALSE;		
	}
	
	public function _canDelete( $chatterID=0, $member=NULL )
	{
		$member = $member ?: \IPS\Member::loggedIn();
		
		if ( $this->_canManage( $member ) || \IPS\chatbox\Rooms\Chat::modPermission( 'delete', NULL, $this ) )
		{
			return TRUE;		
		}
		
		if ( $chatterID > 0 && $member->member_id == $chatterID )
		{
			$perm = $member->_group['g_delete_own_posts'];
			return ( $perm == 1 OR \in_array( 'IPS\chatbox\Rooms\Chat', explode( ',', $perm ) ) ) ? TRUE : FALSE;
		}
		
		return FALSE;		
	}		
	
	public function _canSeeOnline( $member=NULL )
	{
		$member = $member ?: \IPS\Member::loggedIn();
		
		if ( $club = $this->club() )
		{
			$groups = \IPS\Settings::i()->room_clubs_can_see_online;
		}
		else
		{
			$groups = $this->can_see_online;
		}

		if ( $member->inGroup( explode( ",", $groups ) ) )
		{
			return TRUE;		
		}		
		
		return FALSE;
	}
	
	public function _canPlaySong( $member=NULL )
	{
		if ( \IPS\Application::appIsEnabled( 'musicbox' ) && $this->options()['musicbox'] == 1 )
		{
			return TRUE;
		}

		return FALSE;
	}	

	/**
	 * Get the rules
	 *
	 * @return	string
	 */
	public function get__rules()
	{
		if ( \IPS\Member::loggedIn()->language()->checkKeyExists( 'chatbox_room_' . $this->_id . '_rules' ) )
		{
			return \IPS\Member::loggedIn()->language()->get( 'chatbox_room_' . $this->_id . '_rules' );
		}		
	}
	
	/**
	 * Get the rules
	 *
	 * @return	string
	 */
	public function get__description()
	{
		if ( \IPS\Member::loggedIn()->language()->checkKeyExists( 'chatbox_room_' . $this->_id . '_description' ) )
		{
			return \IPS\Member::loggedIn()->language()->get( 'chatbox_room_' . $this->_id . '_description' );
		}
	}	

	/*-------------------------------------------------------------------------*/
	// Get online users
	/*-------------------------------------------------------------------------*/	
	public function online()
	{
		$chatters = $this->getChatters( FALSE );
		$total = \count($chatters);
		
		if ( $total > 0 )
		{
			$last_ping = array_shift($chatters)['last_ping'];
			
			if ( $last_ping <= \IPS\DateTime::create()->sub( new \DateInterval( 'PT300S' ) )->getTimeStamp() )
			{
				$datastoreName = 'chatbox_room_' . $this->id;
				unset( \IPS\Data\Store::i()->$datastoreName );
				return 0;
			}
		}
		
		return $total;
	}
		
	/*-------------------------------------------------------------------------*/
	// Get info from datastore
	/*-------------------------------------------------------------------------*/	
	public function getChatters( $sorting=TRUE )
	{
		$datastoreName = 'chatbox_room_' . $this->id;
		
		$chatters = array();

		if ( isset( \IPS\Data\Store::i()->$datastoreName ) )
		{
			$chatters = \IPS\Data\Store::i()->$datastoreName;
			
			if ( $sorting )
			{
				$myKey = $this->userKey();
				$me = $chatters[ $myKey ];

				unset( $chatters[ $myKey ] );
				
				if ( \count( $chatters ) > 0 )
				{
					uasort($chatters, function ($row1, $row2) {
						return $row2['last_ping'] - $row1['last_ping'];
					});			
					$users = array_merge( array( $myKey => $me ), $chatters );
				}
				else
				{
					$users = array( $myKey => $me );
				}
				return $users;
			}			
		}

		return $chatters;
	}
	
	public function saveChatters( $chatters )
	{
		$datastoreName = 'chatbox_room_' . $this->id;
		\IPS\Data\Store::i()->$datastoreName = $chatters;
		\IPS\Data\Store::i()->chatbox_rooms_update = time();
		
	}
	
	/*-------------------------------------------------------------------------*/
	// Get user info
	/*-------------------------------------------------------------------------*/	
	protected $_userInfo = NULL;
	
	public function userInfo()
	{
		if ( $this->_userInfo !== NULL )
		{
			return $this->_userInfo;
		}
		
		$key = $this->userKey();
		
		if ( \IPS\Member::loggedIn()->member_id > 0 )
		{
			$mem = \IPS\Member::loggedIn();
			$chatterID = $mem->member_id;
			$chatterName = $mem->name;
			$chatterPhoto = $mem->photo;
			$chatterNameFormat = $mem->member_group_id ? \IPS\Member\Group::load( $mem->member_group_id )->formatName( $chatterName ) : $chatterName;					
		}
		else
		{
			$chatterID = 0;
			$chatterName = $this->checkName( str_replace('guest@', '', $key) );
			$chatterPhoto = md5( $chatterName );
			$chatterNameFormat = \IPS\Member\Group::load( \IPS\Settings::i()->guest_group )->formatName( $chatterName );		
		}			
		
		$chatter = array(
			'key'				=> $key,
			'chatterID'			=> $chatterID,
			'chatterName'		=> $chatterName,
			'chatterPhoto'		=> $chatterPhoto,
			'chatterNameFormat'	=> $chatterNameFormat,
			'last_ping'			=> time(),
			'last_chat'			=> 0,
			'ip'				=> \IPS\Request::i()->ipAddress(),
		);
			
		$this->_userInfo = $chatter;
		return $this->_userInfo;
	}
	
	/*-------------------------------------------------------------------------*/
	// Mark message as edited or deleted
	/*-------------------------------------------------------------------------*/
	public function markMsg( $id, $content=NULL )
	{
		$datastoreName = 'chatbox_room_' . $this->id;
		
		if ( ! \IPS\Data\Store::i()->$datastoreName )
		{
			return;
		}
		
		foreach( \IPS\Data\Store::i()->$datastoreName as $key => $val )
		{
			if ( $key != $this->userKey() )
			{
				$val['update'][ $id ] = $content;
			}
			
			$cache[ $key ] = $val;
		}
		
		$this->saveChatters( $cache );
	}
	
	/*-------------------------------------------------------------------------*/
	// Check input name
	/*-------------------------------------------------------------------------*/		
	public function checkName( $txt )
	{
		$name = stripslashes( htmlspecialchars( $txt ) );
		
		$length = mb_strlen($name);

		if ( \is_numeric($name) || !ctype_alnum($name) || !( (3 <= $length) && ($length <= 25)) )
		{
			\IPS\Request::i()->setCookie( 'chatbox_guestName', null );
		
			if ( \IPS\Request::i()->isAjax() )
			{
				\IPS\Output::i()->json( array( 
					'type' 		=> 'errorGuestName', 
					'message' 	=> \IPS\Member::loggedIn()->language()->get( 'chatbox_error_name' ),
				) );
				exit;
			}
			exit;
		}
		else
		{
			\IPS\Request::i()->setCookie( 'chatbox_guestName', $name, \IPS\DateTime::create()->add( new \DateInterval( 'P1D' ) ) );
		}
		
		return $name;
	}	
	
	/*-------------------------------------------------------------------------*/
	// Generate key for user
	/*-------------------------------------------------------------------------*/		
	protected $_userKey = NULL;
	
	public function userKey()
	{	
		if ( $this->_userKey !== NULL )
		{
			return $this->_userKey;
		}
		
		if ( \IPS\Member::loggedIn()->member_id > 0 )
		{
			$key = "user@" . \IPS\Member::loggedIn()->member_id;	
		}
		else
		{
			$key = "guest@" . \IPS\Request::i()->cookie['chatbox_guestName'];	
		}
		
		$this->_userKey = $key;
		return $this->_userKey;
	}

	/*-------------------------------------------------------------------------*/
	// Get announcement
	/*-------------------------------------------------------------------------*/	
	public function announcement()
	{
		$lang = \IPS\Member::loggedIn()->language();
		
		if ( $lang->checkKeyExists( 'chatbox_room_' . $this->id . '_announcement_title' ) && $lang->checkKeyExists( 'chatbox_room_' . $this->id . '_announcement' ) )
		{
			$title = $lang->get( 'chatbox_room_' . $this->id . '_announcement_title' );
			$msg = $lang->get( 'chatbox_room_' . $this->id . '_announcement' );
		}
		return array( $title, $msg );
	}	
	
	/*-------------------------------------------------------------------------*/
	// Add system msg
	/*-------------------------------------------------------------------------*/		
	public function addSystemMsg( $type, $content=NULL )
	{
		if ( $type == 'announcement' && !$this->announcement()[0] )
		{
			$content = 'remove';
		}
		
		$id = \IPS\Db::i()->insert( 'chatbox_messages', array(
			'room'			=> $this->id,
			'member_id'		=> -1,
			'content'		=> $content,
			'system'		=> $type,
			'time'			=> time(),
		) );
		
		$this->cacheLatestID( $id );
	}

	/*-------------------------------------------------------------------------*/
	// Add notification for new content
	/*-------------------------------------------------------------------------*/	
	public function addNotification( $obj )
	{
		# Options
		$opts = json_decode( \IPS\Settings::i()->chatbox_notifications, TRUE );	
		
		# Language
		$lang = \IPS\Member::loggedIn()->language();

		# Is enable?
		$iName = str_replace( '\\', '_', \get_class( $obj ) );
		if ( !$opts[ $iName ] )
		{
			return;
		}
		
		# Content Item
		$item = ( $obj instanceof \IPS\Content\Item ) ? $obj : $obj->item();
		
		# Author
		$author = $obj->author();
		
		# Check container
		if ( isset( $item::$containerNodeClass ) )
		{
			$con = $item->containerWrapper( TRUE );
			$cName = str_replace( '\\', '_', $item::$containerNodeClass );
			if ( $opts[ $cName ] !== '0' && ! \in_array( $con->_id, explode( ',', $opts[ $cName ] ) ) )
			{
				return;
			}
		}
		
		# Format system message
		if( $obj instanceof \IPS\Content\Item && $con )
		{
			$conUrl = " " . $lang->get( 'chatbox_sys_con_in' ) . " " . \IPS\Theme::i()->getTemplate( 'global', 'core', 'global' )->basicUrl( (string) $con->url(), TRUE, $lang->get( $con->_titleLanguageKey ), FALSE );	
		}

		$newUrl = \IPS\Theme::i()->getTemplate( 'global', 'core', 'global' )->basicUrl( (string) $obj->url(), TRUE, $item->mapped( 'title' ), FALSE );
		
		$authorUrl = \IPS\Theme::i()->getTemplate( 'room', 'chatbox', 'front' )->chatterName(
			\IPS\Member\Group::load( $obj->author()->member_group_id )->formatName( $obj->author()->name ), 
			$obj->author()->member_id, 
			(string) $obj->author()->url() 
		);
		
		$data = array( 
			$authorUrl,
			$lang->get( '__indefart_' . $obj::$title ), 
			$conUrl . ": " . $newUrl
		);
		
		$id = \IPS\Db::i()->insert( 'chatbox_messages', array(
			'room'					=> 0,
			'in_multiple_rooms'		=> $opts['rooms'],
			'member_id'				=> -1,
			'content'				=> json_encode( $data ),
			'system'				=> 'content',
			'time'					=> time(),
		) );
		
		// Cache latest id for all rooms
		if ( $opts['rooms'] )
		{
			foreach( explode( ",", $opts['rooms'] ) as $rid )
			{
				$datastoreName = 'chatbox_room_' . $rid . '_latest';
				\IPS\Data\Store::i()->$datastoreName = $id;
			}
		}
	}	

	/**
	 * Format content
	 * @return	string
	 */
	public function formatContent( $txt )
	{
		if ( $this->options()['use_editor'] != 1 && \IPS\Settings::i()->emoji_style == 'twemoji' )
		{
			$txt = \IPS\Application::load('chatbox')->replaceEmoji( $txt );
		}

		return $txt;
	}

	/*-------------------------------------------------------------------------*/
	// Cache latest ID
	/*-------------------------------------------------------------------------*/	
	public function cacheLatestID( $id=0 )
	{
		$cache = array();
		
		if ( isset( \IPS\Data\Store::i()->chatbox_rooms_msg ) )
		{
			$cache = \IPS\Data\Store::i()->chatbox_rooms_msg;
		}
		
		if ( !$cache[ $this->id ] )
		{
			$where[] = array( '( room=? OR FIND_IN_SET( ?, in_multiple_rooms ) )', $this->id, $this->id ); 
			try
			{
				$row = \IPS\Db::i()->select( '*', 'chatbox_messages', $where, 'id DESC' )->first();
				$cache[ $this->id ] = $row['id'];
			}
			catch ( \UnderflowException $e ) 
			{
				$cache[ $this->id ] = 0;
			}
			\IPS\Data\Store::i()->chatbox_rooms_msg = $cache;
		}

		if ( $id > 0 && $cache[ $this->id ] < $id )
		{
			$cache[ $this->id ] = $id;
			\IPS\Data\Store::i()->chatbox_rooms_msg = $cache;
		}
	
		return $cache[ $this->id ];
	}	
	
	/**
	 * Fetch available rooms
	 */
	protected static $_availableRooms = NULL;

	public function availableRooms()
	{
		if ( static::$_availableRooms !== NULL )
		{
			return static::$_availableRooms;
		}	
		
		$rooms = array();
		
		foreach( \IPS\chatbox\Rooms\Room::roots() as $room )
		{
			if ( $room->enabled && $room->_canView() )
			{
				static::$_availableRooms[] = $room;
			}
		}
		
		return static::$_availableRooms;
	}	

	/**
	 * Display chatbox
	 */
	public function cbVars()
	{
		// JS vars
		$opts = $this->options();

		$data = array(
			'roomname' 		=> $this->_title,
			'interval' 		=> (int) $opts['interval'],
			'flood' 		=> (int) $opts['flood'],
			'limit' 		=> (int) $opts['limit'],		
			'length' 		=> (int) $opts['length'],
			'height' 		=> $opts['height'],
			'sort' 			=> $opts['sort'],
			'style' 		=> $opts['style'],
			'giphy' 		=> $opts['giphy'],	
			'guggy' 		=> $opts['guggy'],		
			'allow_img' 	=> $opts['allow_img'] ? true : false,
			'allow_video' 	=> $opts['allow_video'] ? true : false,		
			'autoload' 		=> $opts['autoload'] ? true : false,		
			'canManage' 	=> $this->_canManage() ? true : false,
			'use_editor' 	=> $opts['use_editor'] ? true : false,
			'enter_editor' 	=> $opts['enter_editor'] ? true : false,	
			'soundfile' 	=> $opts['soundfile']? $opts['soundfile'] : NULL,
			'musicbox' 		=> $this->_canPlaySong(),
			'song' 			=> $this->_canPlaySong() && $this->song ? $this->song : NULL,
			'automsg' 		=> $this->autoMsgData() ? $this->autoMsgData() : NULL,
			'alwaysON'		=> \IPS\Request::i()->full == '1' ? true : false,
			'lastchat' 		=> 0,
		);
				
		// Ajax when changing room
		if ( \IPS\Request::i()->isAjax() )
		{
			return $data;
		}
		
		// Return js vars
		$vars[ 'chatbox_' . $this->id ] = $data;
		\IPS\Output::i()->endBodyCode .= \IPS\Theme::i()->getTemplate( 'embed', 'chatbox', 'front' )->chatvars( $vars );
	}
	
	public function show( $widgetConf=NULL, $inGlobal=false, $openFromBar=false )
	{
		# Is Club enabled?
		if ( $this->club() && \IPS\Settings::i()->room_clubs != 1 )
		{
			return;
		}

		# JS Vars
		$this->cbVars();
		
		# Check rules
		if( $this->_canView() && $this->show_rules && \IPS\Request::i()->cookie[ 'chatbox_bypass_rule_' . $this->id ] != 1 )
		{
			$proceed = FALSE;

			$form = new \IPS\Helpers\Form( 'chatbox_rules', 'chatbox_accept_rules' );
			$form->class = 'ipsForm_vertical';
			$form->hiddenValues['proceed'] = TRUE;

			if( $values = $form->values() )
			{
				if( $values['proceed'] )
				{
					$proceed = TRUE;
					\IPS\Request::i()->setCookie( 'chatbox_bypass_rule_' . $this->id, 1, \IPS\DateTime::create()->add( new \DateInterval( 'P1D' ) ) );
				}
			}

			/* If we haven't agreed yet, show the form */
			if( !$proceed )
			{
				$this->showRules = $form;
				
				$ruleHTML = \IPS\Theme::i()->getTemplate( 'room', 'chatbox', 'front' )->chat( $this, $this->options(), $widgetConf, $inGlobal, $openFromBar );
				
				if ( \IPS\Request::i()->full != '1' )
				{
					return $ruleHTML;
				}
				else
				{
					\IPS\Output::i()->sendOutput( \IPS\Theme::i()->getTemplate( 'global', 'core' )->blankTemplate( $ruleHTML ), 200, 'text/html' );
				}
			}
		}
		
		# Auto generate a name for guest
		if ( $this->options()['auto_guest'] == 1 )
		{
			if ( ! \IPS\Member::loggedIn()->member_id && $this->_canView() && ! \IPS\Request::i()->cookie['chatbox_guestName'] )
			{
				$this->checkName( "Guest" . time() );
			}
		}
		
		# Disable sound by default?
		if ( \IPS\Settings::i()->chatbox_sound_disable == 1 && !isset( \IPS\Request::i()->cookie[ 'chatbox_nosound_' . $this->id ] ) )
		{
			\IPS\Request::i()->setCookie( 'chatbox_nosound_' . $this->id, 1, NULL, FALSE );
		}		
		
		# Output
		return \IPS\Theme::i()->getTemplate( 'room', 'chatbox', 'front' )->chat( $this, $this->options(), $widgetConf, $inGlobal, $openFromBar );
	}

	/**
	 * Load Editor
	 */	
	public function editor( $chat=NULL )
	{
		if ( $chat )
		{
			$url = $this->url()->setQueryString( array( 'do' => 'editEditor', 'chatID' => $chat->id ) );
			$btn = 'chatbox_edit';
		}
		else
		{
			$url = $this->url()->setQueryString( 'do', 'postEditor' );
			$btn = 'chatbox_send';
		}
		
		$form = new \IPS\Helpers\Form( 'chatboxEditor', $btn, $url );
		$form->class = 'ipsForm_vertical';
		
		foreach ( \IPS\chatbox\Rooms\Chat::formElements( $chat, $this, $chat ? TRUE : FALSE ) AS $k => $element )
		{
			$form->add( $element );
		}
		
		return $form;		
	}	
	
	
	/**
	 * Show the first room in bar
	 */		
	public function showPopupRoom()
	{
		if ( \IPS\Settings::i()->chatbox_bar_room != 1 )
		{
			return;
		}
		
		if ( \IPS\Settings::i()->chatbox_bar_con == 1 && \IPS\chatbox\Member\Conversation::_canUse() )
		{
			return;
		}
		
		try
		{
			$room = \IPS\chatbox\Rooms\Room::load( self::getMainRoom() );
			
			if ( $room->_canView() )
			{
				return (string) $room->show( NULL, TRUE );
			}
		}
		catch( \OutOfRangeException $e ) 
		{
			\IPS\Request::i()->setCookie( 'chatbox_inRoom', NULL );
		}	
	}
	
	protected static $_getGlobalRooms = NULL;
	
	public function getGlobalRooms()
	{
		if ( static::$_getGlobalRooms !== NULL )
		{
			return static::$_getGlobalRooms;
		}	

		$rooms = self::availableRooms();
		
		if ( \count( $rooms ) > 0 )
		{
			foreach( $rooms as $room )
			{
				if ( \in_array( $room->id, explode(",", \IPS\Settings::i()->chatbox_bar_show_rooms) ) )
				{
					static::$_getGlobalRooms[] = $room;
				}
			}
		}		
		
		return static::$_getGlobalRooms;
	}
	
	protected function getMainRoom()
	{
		if ( \IPS\Request::i()->cookie['chatbox_inRoom'] )
		{
			return \IPS\Request::i()->cookie['chatbox_inRoom'];
		}
		
		$rooms = self::getGlobalRooms();
		
		if ( \count( $rooms ) > 0 )
		{
			return $rooms[0]->id;
		}

		return 0;
	}
	
	
	
	
}